% here we obtain a solution to a 'difficult' cubic equation by perturbing
% off of an 'easy' cubic equation

clear all; close all; clc;

syms x epsil;

% cubic equation to be solved
f = x^3-(4+epsil)*x+2*epsil;

% find a solution close to x=-2
syms a1 a2;
xstar = -2+a1*epsil+a2*epsil^2;
% xstar = +2+a1*epsil+a2*epsil^2;

% substitute xstar into f to get g
g = subs(f,x,xstar)

%%
% collect terms of various of epsil (alternatively you can use taylor)
g = collect(g,epsil)
% setting the coefficients of powers of epsil up to order 2 lets us solve
% for a1 and a2. The two equations we get are:
% 8*a1+4=0 and then
% -12*a1^2-2*a1+16*a2=0

%% %%% %%%
% 
% % alternatively we can pull out the coefficients of powers epsil by
% % expanding g in a taylor series
% 
% % constant term (coefficient of epsil^0) is obtained by setting epsilon=0
% % in g 
coeff0 = subs(g,epsil,0)
% % you will notice that coeff0 = 0. This basically means that when epsil = 0
% % we get xstar = 2. So far so good.
% 
% % coefficient of epsil^1 is obtained by differentiating f once and setting
% % epsil = 0. 
coeff1 = subs(diff(g,epsil),0)
% % you will find that the coefficient was 8*a1 + 4
% 

%%

% % of course, we know that the solution to 8*a1=0 is a1 =0. But let us solve
% % for it symbolically.
a1_sol = solve(coeff1,a1)

%%
% 
% % now we to the next epsil order. to get the coefficient of epsil^2 we find
% % the second derivative of g with respect to epsil and set epsil = 0. Of 
% % course this misses a factor of 2, but we do not care about constant
% % factors as we are going to set the coefficient to zero anyway. If you
% % care, you can divide by 2.
coeff2 = subs(diff(g,epsil,2),0)

%% 
% 
% % now let us substitute our value of a1 into coeff2
coeff2 = subs(coeff2,a1,a1_sol)
% 
%%

% % now let us solve coeff2=0 for a2
a2_sol = solve(coeff2,a2)

% %% Now that we have obtained a solution to the equation upto order 2, let
% % us try compare it to the exact solution for various epsilon


%%
epsil_list = linspace(0,0.7,100)';
xstar_list = -1-(epsil_list+1).^(1/2); % secretly, this is actually an exact solution to the equation close to xstar = -2;
figure(1);
plot(epsil_list,xstar_list,'r'); hold on;

a1_sol = double(a1_sol); % convert these guys to numeric
a2_sol = double(a2_sol);
plot(epsil_list,-2+a1_sol*epsil_list.^1,'g--');
plot(epsil_list,-2+a1_sol*epsil_list.^1+a2_sol*epsil_list.^2,'bo');
legend('exact solution starting at -2', 'one term perturbation approximation', ...
     'two term perturbation approximation');
